structure Lexer =
struct
  datatype token =
    IDENT of string
  | NUMBER of int
  | KWMAIN
  | KWIF
  | KWELSE
  | KWWHILE
  | KWRETURN
  | KWTRUE
  | KWFALSE
  | LBRACE
  | RBRACE
  | LPAREN
  | RPAREN
  | LAND
  | LOR
  | LNOT
  | LESS
  | GREATER
  | EQUAL
  | NOTEQUAL
  | ASSIGN
  | SEMI
  | COMMA
  | PLUS
  | MINUS
  | TIMES
  | DIVIDE


  open Stream

  exception LexicalError

  (* line number, character number *)
  type pos = int * int
  fun pos_to_string (l, c) =
    "line " ^ Int.toString l ^ ", column " ^ Int.toString c

  type t = pos -> (token * pos) front

  type self = {
    main : char stream -> t,
    commentline : char stream -> t,
    commentblock : char stream -> t
  }

  type info = { 
    match : char list,
    len : int,
    start : char stream,
    follow : char stream,
    self : self
  }

  fun action f ({ match, len, follow, self, ... }:info) (l, c) =
     Cons (f (match, len, (l, c)), lazy (fn () => #main self follow (l, c+len)))

  fun simple token ({ len, follow, self, ... }:info) (l, c) =
     Cons ((token, (l, c)), lazy (fn () => #main self follow (l, c+len)))

  structure Arg =
  struct
    type symbol = char
    val ord = Char.ord

    type t = t

    type self = self
    type info = info

    fun eof _ _ = Nil

    fun skip ({ len, follow, self, ... }:info) (l, c) =
      #main self follow (l, c+len)

    fun skip_newline ({ len, follow, self, ... }:info) (l, c) =
      #main self follow (l+1, 0)

    val ident = action (fn (chars, _, pos) => (IDENT (implode chars), pos))

    val number =
      action
        (fn (chars, _, pos) =>
         (case Int.fromString (implode chars) of
            SOME n => (NUMBER n, pos)
          | NONE => raise Bind)
          handle Overflow =>
            (print "Illegal constant at "; print (pos_to_string pos); print ".\n"; raise LexicalError))

    fun error _ pos =
      (print "Lexical error at "; print (pos_to_string pos); print ".\n"; raise LexicalError)

    fun entercommentline ({len, follow, self, ...}:info) (l, c) =
      #commentline self follow (l, c+len)

    fun entercommentblock ({len, follow, self, ...}:info) (l, c) =
      #commentblock self follow (l, c+len)

    fun commentblock_newline ({len, follow, self, ...}:info) (l, c) =
      #commentblock self follow (l+1, 0)

    val kwmain = simple KWMAIN
    val kwif = simple KWIF
    val kwelse = simple KWELSE
    val kwwhile = simple KWWHILE
    val kwreturn = simple KWRETURN
    val kwtrue = simple KWTRUE
    val kwfalse = simple KWFALSE

    val lbrace = simple LBRACE
    val rbrace = simple RBRACE
    val lparen = simple LPAREN
    val rparen = simple RPAREN

    val land = simple LAND
    val lor = simple LOR
    val lnot = simple LNOT

    val less = simple LESS
    val greater = simple GREATER
    val equal = simple EQUAL
    val notequal = simple NOTEQUAL

    val assign = simple ASSIGN
    val semi = simple SEMI
    val comma = simple COMMA

    val plus = simple PLUS
    val minus = simple MINUS
    val times = simple TIMES
    val divide = simple DIVIDE
  end

  structure LexMain =
    MkSystemCLexer
     (structure Streamable = StreamStreamable
      structure Arg = Arg)

  fun lex s = lazy (fn () => LexMain.main s (1, 0))
end
